@@ -1,9 +1,9 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*
- * Copyright (C) 2003      Imendio HB
+ * Copyright (C) 2003-2004 Imendio HB
  * Copyright (C) 2002-2003 Richard Hult <richard@imendio.com>
  * Copyright (C) 2002-2003 Mikael Hallendal <micke@imendio.com>
- * Copyright (C) 2003 Geert-Jan Van den Bogaerde <gvdbogaerde@pandora.be>
+ * Copyright (C) 2003-2004 Geert-Jan Van den Bogaerde <gvdbogaerde@pandora.be>
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -37,12 +37,12 @@
 #include "gossip-app.h"
 #include "gossip-contact-info.h"
 #include "gossip-stock.h"
-#include "disclosure-widget.h"
 #include "gossip-log.h"
 #include "gossip-roster.h"
 #include "gossip-chat.h"
 
 #define d(x)
+
 #define IS_ENTER(v) (v == GDK_Return || v == GDK_ISO_Enter || v == GDK_KP_Enter)
 #define COMPOSING_STOP_TIMEOUT 5
 
@@ -53,19 +53,15 @@ struct _GossipChatPriv {
 
         GtkWidget        *widget;
 	GtkWidget	 *text_view_sw;
-        GtkWidget        *input_entry;
         GtkWidget        *input_text_view;
         GtkWidget        *single_hbox;
-        GtkWidget        *multi_vbox;
-        GtkWidget        *send_multi_button;
         GtkWidget        *subject_entry;
-        GtkWidget        *disclosure;
 
 	GossipChatView   *view;
 
-        GtkTooltips      *tooltips;
-
         GossipRosterItem *item;
+	gchar            *name;
+	
 	gchar            *locked_resource;
 	gchar            *roster_resource;
 
@@ -76,6 +72,7 @@ struct _GossipChatPriv {
 	gchar            *composing_resource; 
 	
 	gboolean          is_online;
+	gboolean          groupchat_priv;
 };
 
 static void            gossip_chat_class_init            (GossipChatClass  *klass);
@@ -99,21 +96,26 @@ static LmHandlerResult chat_message_handler              (LmMessageHandler *hand
 static void            chat_item_presence_updated        (gpointer          not_used,
 							  GossipRosterItem *item,
 							  GossipChat       *chat);
-static gboolean        chat_event_handler                (GossipChat       *chat,
+static void            chat_item_updated                 (gpointer          not_used,
+							  GossipRosterItem *item,
+							  GossipChat       *chat);
+static void            chat_item_removed                 (gpointer          not_used,
+							  GossipRosterItem *item,
+							  GossipChat       *chat);
+static void            chat_item_added			 (gpointer	    not_used,
+							  GossipRosterItem *item,
+							  GossipChat       *chat);
+static void            chat_connected_cb		 (GossipApp	   *app,
+							  GossipChat	   *chat);
+static void            chat_disconnected_cb		 (GossipApp	   *app,
+							  GossipChat	   *chat);
+static gboolean        chat_handle_composing_event       (GossipChat       *chat,
                                                           LmMessage        *m);
 static void            chat_error_dialog                 (GossipChat       *chat,
                                                           const gchar      *msg);
-static void            chat_disclosure_toggled_cb        (GtkToggleButton  *disclosure,
-                                                          GossipChat       *chat);
-static void	       chat_send_multi_clicked_cb	 (GtkWidget	   *button,
-							  GossipChat	   *chat);
-static void            chat_input_activate_cb            (GtkWidget        *entry,
-                                                          GossipChat       *chat);
 static gboolean        chat_input_key_press_event_cb     (GtkWidget        *widget,
                                                           GdkEventKey      *event,
                                                           GossipChat       *chat);
-static void            chat_input_entry_changed_cb       (GtkWidget        *widget,
-                                                          GossipChat       *chat);
 static void            chat_input_text_buffer_changed_cb (GtkTextBuffer    *buffer,
                                                           GossipChat       *chat);
 static gboolean	       chat_text_view_focus_in_event_cb  (GtkWidget        *widget,
@@ -123,14 +125,11 @@ static gboolean	       chat_delete_event_cb		 (GtkWidget	   *widget,
 							  GdkEvent	   *event,
 							  GossipChat       *chat);
 
-/* Called from glade, so it shouldn't be static */
-GtkWidget *     chat_create_disclosure                   (gpointer          data);
-
 enum {
         COMPOSING,
         NEW_MESSAGE,
-        PRESENCE_CHANGED,
-        LAST_SIGNAL
+	NAME_CHANGED,
+	LAST_SIGNAL
 };
 
 static GObjectClass *parent_class = NULL;
@@ -168,6 +167,7 @@ static void
 gossip_chat_class_init (GossipChatClass *klass)
 {
         GObjectClass *object_class = G_OBJECT_CLASS (klass);
+
         parent_class = g_type_class_peek_parent (klass);
 
         object_class->finalize = gossip_chat_finalize;
@@ -193,15 +193,15 @@ gossip_chat_class_init (GossipChatClass *klass)
 			      G_TYPE_NONE,
 			      0);
 
-	chat_signals[PRESENCE_CHANGED] =
-		g_signal_new ("presence-changed",
+	chat_signals[NAME_CHANGED]=
+		g_signal_new ("name-changed",
 			      G_OBJECT_CLASS_TYPE (object_class),
 			      G_SIGNAL_RUN_LAST,
-			      G_STRUCT_OFFSET (GossipChatClass, presence_changed),
+			      G_STRUCT_OFFSET (GossipChatClass, name_changed),
 			      NULL, NULL,
-			      g_cclosure_marshal_VOID__VOID,
+			      g_cclosure_marshal_VOID__POINTER,
 			      G_TYPE_NONE,
-			      0);
+			      1, G_TYPE_POINTER);
 }
 
 static void
@@ -236,11 +236,30 @@ gossip_chat_init (GossipChat *chat)
 			  G_CALLBACK (chat_item_presence_updated),
 			  chat);
 
-	/*g_signal_connect (roster,
+	g_signal_connect (roster,
+			  "item_updated",
+			  G_CALLBACK (chat_item_updated),
+			  chat);
+
+	g_signal_connect (roster,
 			  "item_removed",
-			  G_CALLBACK (chat_item_removed_cb),
+			  G_CALLBACK (chat_item_removed),
+			  chat);
+
+	g_signal_connect (roster,
+			  "item_added",
+			  G_CALLBACK (chat_item_added),
+			  chat);
+
+	g_signal_connect (gossip_app_get (), 
+			  "connected",
+			  G_CALLBACK (chat_connected_cb),
+			  chat);
+
+	g_signal_connect (gossip_app_get (), 
+			  "disconnected",
+			  G_CALLBACK (chat_disconnected_cb),
 			  chat);
-			  */
 }
 
 static void
@@ -263,6 +282,8 @@ gossip_chat_finalize (GObject *object)
 
 	gossip_roster_item_unref (priv->item);
 
+	chat_composing_remove_timeout (chat);
+	
 	g_free (priv->composing_resource);
 	g_free (priv->last_composing_id);
 	
@@ -302,12 +323,7 @@ chat_create_gui (GossipChat *chat)
                                       NULL,
                                       "chat_widget", &priv->widget,
                                       "chat_view_sw", &priv->text_view_sw,
-                                      "input_entry", &priv->input_entry,
                                       "input_textview", &priv->input_text_view,
-                                      "single_hbox", &priv->single_hbox,
-                                      "multi_vbox", &priv->multi_vbox,
-                                      "disclosure", &priv->disclosure,
-                                      "send_multi_button", &priv->send_multi_button,
                                       NULL);
 
 	priv->view = gossip_chat_view_new ();
@@ -321,32 +337,10 @@ chat_create_gui (GossipChat *chat)
 
         roster = gossip_app_get_roster ();
 
-	priv->tooltips = gtk_tooltips_new ();
-
-        g_signal_connect (priv->disclosure,
-                          "toggled",
-                          G_CALLBACK (chat_disclosure_toggled_cb),
-                          chat);
-	g_signal_connect (priv->send_multi_button,
-			  "clicked",
-			  G_CALLBACK (chat_send_multi_clicked_cb),
-			  chat);
-        g_signal_connect (chat->priv->input_entry,
-                          "activate",
-                          G_CALLBACK (chat_input_activate_cb),
-                          chat);
-        g_signal_connect (priv->input_entry,
-                          "changed",
-                          G_CALLBACK (chat_input_entry_changed_cb),
-                          chat);
 	g_signal_connect (priv->input_text_view,
 			  "key_press_event",
 			  G_CALLBACK (chat_input_key_press_event_cb),
 			  chat);
-        g_signal_connect (priv->input_entry,
-                          "key_press_event",
-                          G_CALLBACK (chat_input_key_press_event_cb),
-                          chat);
 
         buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (priv->input_text_view));
         g_signal_connect (buffer,
@@ -364,7 +358,7 @@ chat_create_gui (GossipChat *chat)
 
 	gossip_chat_view_set_margin (priv->view, 3);
 
-	gtk_widget_grab_focus (priv->input_entry);
+	gtk_widget_grab_focus (priv->input_text_view);
 }
 
 static void
@@ -374,6 +368,10 @@ chat_update_locked_resource (GossipChat *chat)
 	GossipJID      *jid;
 	const gchar    *roster_resource;
 
+	if (priv->groupchat_priv) {
+		return;
+	}
+	
 	if (gossip_roster_item_is_offline (priv->item)) {
 		g_free (priv->roster_resource);
 		priv->roster_resource = NULL;
@@ -457,16 +455,14 @@ chat_send (GossipChat  *chat,
 
 	priv = chat->priv;
 
+	gossip_app_force_non_away ();
+
         if (msg == NULL || msg[0] == '\0') {
                 return;
         }
 
         if (g_ascii_strcasecmp (msg, "/clear") == 0) {
-                GtkTextBuffer *buffer;
-
-                buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (priv->view));
-                gtk_text_buffer_set_text (buffer, "", 1);
-
+		gossip_chat_view_clear (priv->view);
                 return;
         }
 
@@ -476,7 +472,7 @@ chat_send (GossipChat  *chat,
 
         gossip_chat_view_append_chat_message (priv->view,
                                               NULL,
-                                              gossip_app_get_username (),
+					      gossip_app_get_username (),
                                               nick,
                                               msg);
 
@@ -640,6 +636,33 @@ chat_composing_stop_timeout_cb (GossipChat *chat)
         return FALSE;
 }
 
+static gboolean
+chat_should_play_sound (GossipChat *chat)
+{
+	GossipChatPriv *priv = chat->priv;
+	GtkWidget      *toplevel;
+	gboolean        play = TRUE;
+
+	/* Play sounds if the window is not focused. */
+	
+	if (!priv->window) {
+		return TRUE;
+	}
+
+	toplevel = gossip_chat_window_get_dialog (priv->window);
+
+	/* The has-toplevel-focus property is new in GTK 2.2 so if we don't find it, we
+	 * pretend that the window doesn't have focus => always play sound.
+	 */
+	if (g_object_class_find_property (G_OBJECT_GET_CLASS (toplevel),
+					  "has-toplevel-focus")) {
+		g_object_get (toplevel, "has-toplevel-focus", &play, NULL);
+		play = !play;
+	}
+
+	return play;
+}
+
 static LmHandlerResult
 chat_message_handler (LmMessageHandler *handler,
                       LmConnection     *connection,
@@ -665,21 +688,30 @@ chat_message_handler (LmMessageHandler *handler,
         from_jid = gossip_jid_new (from);
 	jid = gossip_roster_item_get_jid (priv->item);
 
+	if (lm_message_get_sub_type (m) == LM_MESSAGE_SUB_TYPE_GROUPCHAT) {
+		d(g_print ("GROUP CHAT!\n"));
+		gossip_jid_unref (from_jid);
+		
+                return LM_HANDLER_RESULT_ALLOW_MORE_HANDLERS;
+	}
+	
         d(g_print ("Incoming message:: '%s' ?= '%s'\n",
-                   gossip_jid_get_without_resource (from_jid),
-                   gossip_jid_get_without_resource (jid)));
+                   gossip_jid_get_full (from_jid),
+                   gossip_jid_get_full (jid)));
 
-	d(g_print ("from: '%s'\n", gossip_jid_get_full (from_jid)));
+        if ((!priv->groupchat_priv && !gossip_jid_equals_without_resource (from_jid, jid)) ||
 
-        if (!gossip_jid_equals_without_resource (from_jid, jid)) {
-                gossip_jid_unref (from_jid);
+	    (priv->groupchat_priv && !gossip_jid_equals (from_jid, jid))) {
+		d(g_print ("GROUP CHAT2!\n"));
+		gossip_jid_unref (from_jid);
+		
                 return LM_HANDLER_RESULT_ALLOW_MORE_HANDLERS;
         }
 
         type = lm_message_get_sub_type (m);
 
         if (type == LM_MESSAGE_SUB_TYPE_ERROR) {
-                gchar     *tmp, *str, *msg;
+                gchar *tmp, *str, *msg;
 
                 tmp = g_strdup_printf ("<b>%s</b>", from);
                 str = g_strdup_printf (_("An error occurred when chatting with %s."), tmp);
@@ -713,7 +745,7 @@ chat_message_handler (LmMessageHandler *handler,
 	
 	gossip_jid_unref (from_jid);
 
-        if (chat_event_handler (chat, m)) {
+        if (chat_handle_composing_event (chat, m)) {
                 return LM_HANDLER_RESULT_REMOVE_MESSAGE;
         }
 
@@ -733,12 +765,16 @@ chat_message_handler (LmMessageHandler *handler,
 
         gossip_chat_view_append_chat_message (chat->priv->view,
                                               timestamp,
-                                              gossip_app_get_username (),
+					      NULL,
                                               gossip_roster_item_get_name (priv->item),
                                               body);
 
 	g_signal_emit (chat, chat_signals[NEW_MESSAGE], 0);
 	
+	if (chat_should_play_sound (chat)) {
+		gossip_sound_play (GOSSIP_SOUND_CHAT);
+	}
+
 	if (priv->window == NULL) {
 		return LM_HANDLER_RESULT_ALLOW_MORE_HANDLERS;
 	}
@@ -795,12 +831,111 @@ chat_item_presence_updated (gpointer          not_used,
 		}
 		priv->is_online = TRUE;
 	}
+}
 
-	g_signal_emit (chat, chat_signals[PRESENCE_CHANGED], 0);
+static void
+chat_item_updated (gpointer          not_used,
+		   GossipRosterItem *item,
+		   GossipChat       *chat)
+{
+	GossipChatPriv *priv;
+
+	g_return_if_fail (GOSSIP_IS_CHAT (chat));
+	g_return_if_fail (item != NULL);
+	
+	priv = chat->priv;
+
+	if (item != priv->item) {
+		return;
+	}
+
+	if (strcmp (priv->name, gossip_roster_item_get_name (item)) != 0) {
+		g_free (priv->name);
+		priv->name = g_strdup (gossip_roster_item_get_name (item));
+		g_signal_emit (chat, chat_signals[NAME_CHANGED], 0, priv->name);
+	}
+}
+
+static void 
+chat_item_removed (gpointer          not_used,
+		   GossipRosterItem *item,
+		   GossipChat       *chat)
+{
+	GossipChatPriv *priv;
+
+	g_return_if_fail (GOSSIP_IS_CHAT (chat));
+	g_return_if_fail (item != NULL);
+	
+	priv = chat->priv;
+
+	if (item != priv->item) {
+		return;
+	}
+}
+
+static void
+chat_item_added (gpointer          not_user,
+		 GossipRosterItem *item,
+		 GossipChat       *chat)
+{
+	GossipChatPriv *priv;
+	GossipJID      *new_jid;
+	GossipJID      *old_jid;
+
+	g_return_if_fail (GOSSIP_IS_CHAT (chat));
+	g_return_if_fail (item != NULL);
+
+	priv = chat->priv;
+	
+	new_jid = gossip_roster_item_get_jid (item);
+	old_jid = gossip_roster_item_get_jid (priv->item);
+
+	if (!gossip_jid_equals_without_resource (old_jid, new_jid))
+		return;
+
+	gossip_roster_item_unref (priv->item);
+	priv->item = item;
+	gossip_roster_item_ref (item);
+}
+
+static void
+chat_connected_cb (GossipApp  *app,
+		   GossipChat *chat)
+{
+	GossipChatPriv *priv;
+
+	g_return_if_fail (GOSSIP_IS_CHAT (chat));
+
+	priv = chat->priv;
+
+	gtk_widget_set_sensitive (priv->input_text_view, TRUE);
+
+	gossip_chat_view_append_event_msg (priv->view, _("Connected"), TRUE);
+}
+
+static void
+chat_disconnected_cb (GossipApp  *app,
+		      GossipChat *chat)
+{
+	GossipChatPriv *priv;
+
+	g_return_if_fail (GOSSIP_IS_CHAT (chat));
+
+	priv = chat->priv;
+
+	gtk_widget_set_sensitive (priv->input_text_view, FALSE);
+
+	gossip_chat_view_append_event_msg (priv->view, _("Disconnected"), TRUE);
+
+	priv->send_composing_events = FALSE;
+	chat_composing_remove_timeout (chat);
+
+	g_free (priv->composing_resource);
+	priv->composing_resource = NULL;
 }
 
 static gboolean
-chat_event_handler (GossipChat *chat, LmMessage *m)
+chat_handle_composing_event (GossipChat *chat, LmMessage *m)
 {
 	GossipChatPriv *priv = chat->priv;
         LmMessageNode  *x;
@@ -820,13 +955,13 @@ chat_event_handler (GossipChat *chat, LmMessage *m)
         }
 
         if (lm_message_node_get_child (m->node, "body")) {
-                if (lm_message_node_get_child (x, "composing")) {
+                if (priv->is_online && lm_message_node_get_child (x, "composing")) {
                         /* Handle request for composing events. */
 			priv->send_composing_events = TRUE;
-			
+
 			from = lm_message_node_get_attribute (m->node, "from");
 			jid = gossip_jid_new (from);
-			
+
 			g_free (priv->composing_resource);
 			priv->composing_resource =
 				g_strdup (gossip_jid_get_resource (jid));
@@ -858,12 +993,16 @@ static void
 chat_error_dialog (GossipChat  *chat, 
                    const gchar *msg)
 {
-        GtkWidget *window;
+        GtkWindow *window = NULL;
         GtkWidget *dialog;
 
-        window = gossip_chat_window_get_dialog (chat->priv->window);
-        
-        dialog = gtk_message_dialog_new (GTK_WINDOW (window),
+        g_return_if_fail (chat != NULL);
+
+        if (chat->priv->window) {
+                window = GTK_WINDOW (gossip_chat_window_get_dialog (chat->priv->window));
+        }
+
+        dialog = gtk_message_dialog_new (window,
                                          GTK_DIALOG_DESTROY_WITH_PARENT,
                                          GTK_MESSAGE_ERROR,
                                          GTK_BUTTONS_CLOSE,
@@ -880,36 +1019,7 @@ chat_error_dialog (GossipChat  *chat,
 }
 
 static void
-chat_disclosure_toggled_cb (GtkToggleButton *disclosure,
-                            GossipChat      *chat)
-{
-        GtkTextBuffer *buffer;
-        GtkTextIter    start, end;
-        const gchar   *const_str;
-        gchar         *str;
-
-        buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (chat->priv->input_text_view));
-
-        if (gtk_toggle_button_get_active (disclosure)) {
-                gtk_widget_show (chat->priv->multi_vbox);
-                gtk_widget_hide (chat->priv->single_hbox);
-
-                const_str = gtk_entry_get_text (GTK_ENTRY (chat->priv->input_entry));
-                gtk_text_buffer_set_text (buffer, const_str, -1);
-        } else {
-                gtk_widget_show (chat->priv->single_hbox);
-                gtk_widget_hide (chat->priv->multi_vbox);
-
-                gtk_text_buffer_get_bounds (buffer, &start, &end);
-                str = gtk_text_buffer_get_text (buffer, &start, &end, FALSE);
-                gtk_entry_set_text (GTK_ENTRY (chat->priv->input_entry), str);
-                g_free (str);
-        }
-}
-
-static void
-chat_send_multi_clicked_cb (GtkWidget  *button,
-			    GossipChat *chat)
+chat_input_text_view_send (GossipChat *chat)
 {
 	GtkTextBuffer *buffer;
 	GtkTextIter    start, end;
@@ -932,20 +1042,16 @@ chat_send_multi_clicked_cb (GtkWidget  *button,
 	g_free (msg);
 }
 
-static void
-chat_input_activate_cb (GtkWidget  *entry,
-                        GossipChat *chat)
+void
+gossip_chat_scroll_down (GossipChat *chat)
 {
-        gchar *msg;
+	GossipChatPriv *priv;
+	
+	g_return_if_fail (GOSSIP_IS_CHAT (chat));
 
-        msg = gtk_editable_get_chars (GTK_EDITABLE (chat->priv->input_entry), 0, -1);
+	priv = chat->priv;
 
-        /* Clear the input field */
-        gtk_entry_set_text (GTK_ENTRY (chat->priv->input_entry), "");
-
-        chat_send (chat, msg);
-
-        g_free (msg);
+	gossip_chat_view_scroll_down (priv->view);
 }
 
 static gboolean
@@ -953,47 +1059,46 @@ chat_input_key_press_event_cb (GtkWidget   *widget,
                                GdkEventKey *event,
                                GossipChat  *chat)
 {
-        /* Catch ctrl-enter */
-        if ((event->state && GDK_CONTROL_MASK) && IS_ENTER (event->keyval)) {
-                if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (chat->priv->disclosure))) {
-                        gtk_widget_activate (chat->priv->send_multi_button);
-                } else {
-                        gtk_widget_activate (chat->priv->input_entry);
-                }
+	GossipChatPriv *priv = chat->priv;
+	GtkAdjustment  *adj;
+	gdouble         val;
+		
+	/* Catch enter but not ctrl/shift-enter */
+	if (IS_ENTER (event->keyval) && !(event->state & GDK_CONTROL_MASK) &&
+	    !(event->state & GDK_SHIFT_MASK)) {
+		chat_input_text_view_send (chat);
+		
+		return TRUE;
+	}
+	if (IS_ENTER (event->keyval) && (event->state & GDK_CONTROL_MASK)) {
+		/* Do nothing for ctrl-enter. */
+		return TRUE;
+	}
+	if (IS_ENTER (event->keyval) && (event->state & GDK_SHIFT_MASK)) {
+		/* Newline for shift-enter. */
+		return FALSE;
+	}
+	else if (event->keyval == GDK_Page_Up) {
+		adj = gtk_scrolled_window_get_vadjustment (GTK_SCROLLED_WINDOW (priv->text_view_sw));
+		gtk_adjustment_set_value (adj, adj->value - adj->page_size);
+		
+		return TRUE;
+	}
+	else if (event->keyval == GDK_Page_Down) {
+		adj = gtk_scrolled_window_get_vadjustment (GTK_SCROLLED_WINDOW (priv->text_view_sw));
+		val = MIN (adj->value + adj->page_size, adj->upper - adj->page_size);  
+		gtk_adjustment_set_value (adj, val);
 
-                return TRUE;
-        }
-
-        return FALSE;
-}
-
-static void
-chat_input_entry_changed_cb (GtkWidget  *widget,
-                             GossipChat *chat)
-{
-        const gchar *str;
-
-        if (chat->priv->send_composing_events) {
-                str = gtk_entry_get_text (GTK_ENTRY (widget));
-
-                if (strlen (str) == 0) {
-                        chat_composing_stop (chat);
-                } else {
-                        chat_composing_start (chat);
-                }
-        }
+		return TRUE;
+	}
+		
+	return FALSE;
 }
 
 static void
 chat_input_text_buffer_changed_cb (GtkTextBuffer *buffer,
                                    GossipChat    *chat)
 {
-        if (gtk_text_buffer_get_line_count (buffer) > 1) {
-                gtk_widget_set_sensitive (chat->priv->disclosure, FALSE);
-        } else {
-                gtk_widget_set_sensitive (chat->priv->disclosure, TRUE);
-        }
-
         if (chat->priv->send_composing_events) {
                 if (gtk_text_buffer_get_char_count (buffer) == 0) {
                         chat_composing_stop (chat);
@@ -1008,15 +1113,10 @@ chat_text_view_focus_in_event_cb (GtkWidget  *widget,
 				  GdkEvent   *event,
 				  GossipChat *chat)
 {
-	gint pos;
-
-	pos = gtk_editable_get_position (GTK_EDITABLE (chat->priv->input_entry));
-
-	gtk_widget_grab_focus (chat->priv->input_entry);
-	gtk_editable_select_region (GTK_EDITABLE (chat->priv->input_entry), 0, 0);
-
-	gtk_editable_set_position (GTK_EDITABLE (chat->priv->input_entry), pos);
+	GossipChatPriv *priv = chat->priv;
 
+	gtk_widget_grab_focus (priv->input_text_view);
+	
 	return TRUE;
 }
 
@@ -1028,20 +1128,8 @@ chat_delete_event_cb (GtkWidget  *widget,
 	return TRUE;
 }
 
-GtkWidget *
-chat_create_disclosure (gpointer data)
-{
-	GtkWidget *widget;
-
-	widget = cddb_disclosure_new (NULL, NULL);
-
-	gtk_widget_show (widget);
-
-	return widget;
-}
-
 GossipChat *
-gossip_chat_get_for_item (GossipRosterItem *item)
+gossip_chat_get_for_item (GossipRosterItem *item, gboolean create)
 {
 	GossipChat     *chat;
 	GossipChatPriv *priv;
@@ -1055,12 +1143,17 @@ gossip_chat_get_for_item (GossipRosterItem *item)
 	if (chat) {
 		return chat;
 	}
+
+	if (!create)
+		return NULL;
 	
 	chat = g_object_new (GOSSIP_TYPE_CHAT, NULL);
 	g_hash_table_insert (chats, gossip_jid_ref (jid), chat);
 	
 	priv = chat->priv;
 	priv->item = gossip_roster_item_ref (item);
+	priv->name = g_strdup (gossip_roster_item_get_name (item));
+	priv->groupchat_priv = FALSE;
 
 	if (gossip_roster_item_is_offline (priv->item)) {
 		priv->is_online = FALSE;
@@ -1089,6 +1182,9 @@ gossip_chat_get_for_group_chat (GossipRosterItem *item)
 	chat = g_object_new (GOSSIP_TYPE_CHAT, NULL);
 	priv = chat->priv;
 	priv->item = gossip_roster_item_ref (item);
+	priv->groupchat_priv = TRUE;
+
+	priv->locked_resource = g_strdup (gossip_jid_get_resource (jid));
 	
 	if (gossip_roster_item_is_offline (priv->item)) {
 		priv->is_online = FALSE;
@@ -1099,6 +1195,31 @@ gossip_chat_get_for_group_chat (GossipRosterItem *item)
 	return chat;
 }
 
+gchar *
+gossip_chat_get_history (GossipChat *chat, gint lines)
+{
+	GossipChatPriv *priv;
+	GtkTextBuffer  *buffer;
+	GtkTextIter     start, end;
+
+	priv = chat->priv;
+
+	buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (priv->view));
+
+	gtk_text_buffer_get_bounds (buffer, &start, &end);
+
+	if (lines != -1) {
+		GtkTextIter tmp;
+
+		tmp = end;
+		if (gtk_text_iter_backward_lines (&tmp, lines)) {
+			start = tmp;
+		}
+	}
+
+	return gtk_text_buffer_get_text (buffer, &start, &end, FALSE);
+}
+
 void
 gossip_chat_append_message (GossipChat *chat, LmMessage *m)
 {
@@ -1137,7 +1258,7 @@ gossip_chat_handle_message (LmMessage *m)
 		/* The existing message handler will catch it. */
 		result = LM_HANDLER_RESULT_ALLOW_MORE_HANDLERS;
 	} else {
-		chat = gossip_chat_get_for_item (item);
+		chat = gossip_chat_get_for_item (item, TRUE);
 		/* chat = g_object_new (GOSSIP_TYPE_CHAT,
 				     "jid", jid, NULL); */
 		gossip_chat_append_message (chat, m);
@@ -1166,7 +1287,7 @@ gossip_chat_present (GossipChat *chat)
         gossip_chat_window_switch_to_chat (priv->window, chat);
         gtk_window_present (GTK_WINDOW (gossip_chat_window_get_dialog (priv->window)));
 
-	gtk_widget_grab_focus (priv->input_entry);
+	gtk_widget_grab_focus (priv->input_text_view);
 }
 
 GtkWidget *
@@ -1198,3 +1319,16 @@ gossip_chat_get_window (GossipChat *chat)
 {
 	return chat->priv->window;
 }
+
+void
+gossip_chat_clear (GossipChat *chat)
+{
+	GossipChatPriv *priv;
+	
+	g_return_if_fail (GOSSIP_IS_CHAT (chat));
+
+	priv = chat->priv;
+
+	gossip_chat_view_clear (priv->view);
+}
+
