@@ -1,7 +1,7 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*
- * Copyright (C) 2003 Imendio HB
- * Copyright (C) 2003 Johan Wallenborg <johan.wallenborg@fishpins.se>
+ * Copyright (C) 2003-2004 Imendio HB
+ * Copyright (C) 2003      Johan Wallenborg <johan.wallenborg@fishpins.se>
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -73,13 +73,17 @@ static gchar *
 log_get_filename (GossipJID *jid, const gchar *suffix)
 {
 	gchar *tmp;
+	gchar *ret;
 	
 	tmp = g_build_filename (g_get_home_dir (),
 				".gnome2", "Gossip", "logs",
 				gossip_jid_get_without_resource (jid),
 				NULL);
 
-	return g_strconcat (tmp, suffix, NULL);
+	ret = g_strconcat (tmp, suffix, NULL);
+	g_free (tmp);
+
+	return ret;
 }
 
 static gchar *
@@ -104,6 +108,79 @@ log_get_timestamp (LmMessage *msg)
 	return g_strdup (buf);
 }
 
+static gchar *
+log_urlify (const gchar *msg)
+{
+	gint     num_matches, i;
+	GArray  *start, *end;
+	GString *ret;
+	gchar   *esc;
+
+	ret = g_string_new (NULL);
+	
+	start = g_array_new (FALSE, FALSE, sizeof (gint));
+	end = g_array_new (FALSE, FALSE, sizeof (gint));
+	
+	num_matches = gossip_utils_url_regex_match (msg, start, end);
+
+	if (num_matches == 0) {
+		esc = g_markup_escape_text (msg, -1);
+		g_string_append (ret, esc);
+		g_free (esc);
+	} else {
+		gint   last = 0;
+		gint   s = 0, e = 0;
+		gchar *tmp;
+
+		for (i = 0; i < num_matches; i++) {
+
+			s = g_array_index (start, gint, i);
+			e = g_array_index (end, gint, i);
+
+			if (s > last) {
+				tmp = gossip_utils_substring (msg, last, s);
+				esc = g_markup_escape_text (tmp, -1);
+				g_string_append (ret, esc);
+				g_free (tmp);
+				g_free (esc);
+			}
+
+			tmp = gossip_utils_substring (msg, s, e);
+
+			g_string_append (ret, "<a href=\"");
+
+			esc = g_markup_escape_text (tmp, -1);
+			g_string_append (ret, esc);
+			g_free (esc);
+
+			g_string_append (ret, "\">");
+
+			esc = g_markup_escape_text (tmp, -1);
+			g_string_append (ret, esc);
+			g_free (esc);
+
+			g_string_append (ret, "</a>");
+			
+			last = e;
+
+			g_free (tmp);
+		}
+
+		if (e < strlen (msg)) {
+			tmp = gossip_utils_substring (msg, e, strlen (msg));
+			esc = g_markup_escape_text (tmp, -1);
+			g_string_append (ret, esc);
+			g_free (tmp);
+			g_free (esc);
+		}
+	}
+
+	g_array_free (start, TRUE);
+	g_array_free (end, TRUE);
+	
+	return g_string_free (ret, FALSE);
+}
+
 void
 gossip_log_message (LmMessage *msg, gboolean incoming)
 {
@@ -170,7 +247,11 @@ gossip_log_message (LmMessage *msg, gboolean incoming)
 
 	node = lm_message_node_get_child (msg->node, "body");
 	if (node) {
-		body = g_markup_escape_text (node->value, -1);
+		gchar *tmp;
+
+		tmp = log_urlify (node->value);
+		body = g_strdup (tmp); //g_markup_escape_text (tmp, -1);
+		g_free (tmp);
 	} else {
 		body = g_strdup ("");
 	}
